"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from requests_toolbelt.multipart.encoder import MultipartEncoder
from typing import Dict
from .serviceconnector import _Client, ServiceConnector
from .message import Message
from .utils import get_logger
from .camel import CamelResource
from .utils import raise_for_status_with_detail

log = get_logger(__name__)


class ActionClient(_Client):
    """
    A client for the Cortex Actions API.
    """

    URIs = {
        'deploy':       'actions',
        'invoke':       'actions/{action_name}/invoke',
        'delete':       'actions/{action_name}',
        'logs':         'actions/{action_name}/logs',
        'job_stats':    'jobs/{action_name}/stats',
        'tasks':        'jobs/{action_name}/tasks',
        'task_logs':    'jobs/{action_name}/tasks/{task_id}/logs',
        'task_status':  'jobs/{action_name}/tasks/{task_id}/status',
        'task_delete':  'jobs/{action_name}/tasks/{task_id}',
        'send_message': 'actions/send-message',
    }

    def deploy_action(self, name, kind: str, docker: str, code = '', action_type = None, **kwargs):
        """
        Deploys an action.

        :param name: The resource name of the action.
        :param kind: The action kind (only for functions) - python:2 or python:3
        :param docker: The Docker image to use (optional for functions, required for **jobs** and **daemons**)
        :param code: The action code to deploy; expects a file-like object and zip compressed contents.
        :param action_type: The type of action workload: function, daemon, job (default: function).
        :return: The action deployment result.
        """
        action = {'name': name}

        if docker:
            action['docker'] = docker

        if kind:
            action['kind'] = kind

        if code:
            try:
                code_file = open(code, 'rb')
            except TypeError:
                code_file = code

            action['code'] = ('code.zip', code_file, 'application/zip')

        if kwargs.get('command'):
            action['command'] = kwargs.get('command')

        if kwargs.get('port'):
            action['port'] = kwargs.get('port')

        if kwargs.get('environment'):
            action['environment'] = kwargs.get('environment')

        if kwargs.get('vcpus'):
            action['vcpus'] = str(kwargs.get('vcpus'))

        if kwargs.get('memory'):
            action['memory'] = str(kwargs.get('memory'))

        if kwargs.get('backend_type'):
            action['backend_type'] = kwargs.get('backend_type')

        uri = self.URIs['deploy']
        if action_type is not None:
            uri = '{}?actionType={}'.format(uri, action_type)

        m = MultipartEncoder(action)
        r = self._serviceconnector.request(method='POST', uri=uri, body=m, headers={'Content-type': m.content_type})
        raise_for_status_with_detail(r)
        return r.json()

    def invoke_action(self, action_name, params: Dict[str, object]) -> Dict[str, object]:
        """
        Invokes an action.

        :param action_name: The name of the action to invoke.
        :param params: The body params to send the action.

        :return: The result of calling the action.
        """
        uri = self.URIs['invoke'].format(action_name=action_name)
        return self._post_json_with_retry(uri, params)

    def get_logs(self, action_name) -> Dict[str, object]:
        """
        Gets the most recent logs for an action.

        :param action_name: The action name to retrieve logs from.
        :return: The most recent logs for the requested action.
        """
        uri = self.URIs['logs'].format(action_name=action_name)
        return self._get_json(uri) or {}

    def delete_action(self, action_name, action_type=None):
        """
        Deletes an action.

        :param action_name: The name of the action to delete.
        :return: The status of the deletion.
        """
        uri = self.URIs['delete'].format(action_name=action_name)
        if action_type:
            uri = '{}?actionType={}'.format(uri, action_type)

        r = self._serviceconnector.request(method='DELETE', uri=uri)
        raise_for_status_with_detail(r)

        return r.json()

    def get_task_status(self, action_name, task_id):
        """
        Gets the status for an action task (for job type actions only).

        :param action_name: The name of the job action for the task.
        :param task_id: The task id to get status for.
        :return: Task status.
        """
        uri = self.URIs['task_status'].format(action_name=action_name, task_id=task_id)
        r = self._serviceconnector.request(method='GET', uri=uri)
        raise_for_status_with_detail(r)
        return r.text


    def get_task_logs(self, action_name, task_id):
        """
        Gets the logs for an action task (for job type actions only).

        :param action_name: The name of the job action for the task.
        :param task_id: The task id to get logs for.
        :return: Task logs.
        """
        uri = self.URIs['task_logs'].format(action_name=action_name, task_id=task_id)
        return self._get_json(uri)

    def get_job_stats(self, action_name):
        """
        Gets execution stats for tasks associated with action `action_name` of type job.

        :param action_name: The name of the job action.
        :return: A `dict` with stats information.
        """
        uri = self.URIs['job_stats'].format(action_name=action_name)
        return self._get_json(uri)

    def get_tasks(self, action_name):
        """
        Gets list of tasks associated with the give job `action_name`.

        :param action_name: The name of the job action.
        :return: A `dict` with tasks.
        """
        uri = self.URIs['tasks'].format(action_name=action_name)
        return self._get_json(uri)

    def delete_task(self, action_name, task_id):
        """
        Deletes a task associated with a job action.

        :param action_name: The name of the job action.
        :param task_id: The ID of the task to delete.
        :return: Deletion status.
        """
        uri = self.URIs['task_delete'].format(action_name=action_name, task_id=task_id)
        r = self._serviceconnector.request(method='DELETE', uri=uri)
        raise_for_status_with_detail(r)
        return r.json()

    def send_message(self, message):
        """
        Send Synapse Message

        :param message: Input message
        :return: status of message
        """
        uri = self.URIs['send_message']
        return self._post_json(uri, message)


class BuilderClient(_Client):
    """
    A client for the Cortex Docker builder service REST API.
    """

    URIs = {'action': 'builder/action'}

    def post_job(self, build_dir, image_tag):
        """
        Submit a build job request
        """
        body = {
            'imageTag' : image_tag,
            'buildContext': build_dir
        }
        return self._post_json(self.URIs['action'], body)

    def get_job(self, jobid):
        """
        Gets targets for the agent.
        """
        return self._get_json(self.URIs['action'] + '/{}'.format(jobid))

    def get_job_logs(self, jobid):
        """
        Gets job logs.
        """
        url = '/'.join([self.URIs['action'], jobid, 'logs'])
        return self._get(url, stream=True)


class Action(CamelResource):
    """
    A Cortex Action. Actions are the computional part of a skill.
    """
    def __init__(self, action, connector: ServiceConnector):
        super().__init__(action, True)
        self._connector = connector

    def invoke(self, message: Message, timeout=30):
        """
        Invoke an action.

        :param message: A Message to send as input to the action.
        :return: A response Message.
        """
        action_client = ActionClient(self._connector.url, self._connector.version, self._connector.token)
        return Message(action_client.invoke_action(self.name, message.to_params()))

    def get_deployment_status(self):
        """
        Get the deployment status of an action.
        """
        a = Action.get_action(self.name, self._connector)
        return a.deploymentStatus

    def delete(self):
        """
        Delete an action.
        """
        action_client = ActionClient(self._connector.url, self._connector.version, self._connector.token)
        return action_client.delete_action(self.name, self.type)

    def get_task_status(self, task_id):
        """
        Get the status for a job task (a task is a particular invocation of the action type: job).

        :param task_id: The identifier for the task.
        """
        action_client = ActionClient(self._connector.url, self._connector.version, self._connector.token)
        return action_client.get_task_status(self.name, task_id)

    def get_task_logs(self, task_id):
        """
        Get the logs for a task.

        :param task_id: The identifier for the task.
        """
        action_client = ActionClient(self._connector.url, self._connector.version, self._connector.token)
        return action_client.get_task_logs(self.name, task_id)

    def _repr_(self):
        return "Action(name={name}, version={version}, type={type}, kind={kind})"\
            .format(name=self.name, version=self.version, kind=self.kind, type=self.type)

    def _repr_html_(self):
        template = """
        <table>
            <thead>
                <tr>
                    <th>Name</th>
                    <th>Version</th>
                    <th>Type</th>
                    <th>Kind</th>
                    <th>Image</th>
                    <th>Deployment Status</th>
                </tr>
            </thead>
            <tbody>
                <td>{name}</td>
                <td>{version}</td>
                <td>{type}</td>
                <td>{kind}</td>
                <td>{image}</td>
                <td>{status}</td>
            </tbody>
        </table>
        """

        return template.format(name=self.name, version=self.version, type=self.type,
                               kind=self.kind, image=self.image, status=self.get_deployment_status())


    @staticmethod
    def get_action(name: str, connector: ServiceConnector):
        """
        Fetches an Action to work with.

        :param name: The name of the Action to retrieve.

        :return: An Action object.
        """
        uri = 'actions/{name}'.format(name=name)
        log.debug('Getting action using URI: %s' % uri)
        r = connector.request('GET', uri)
        raise_for_status_with_detail(r)
        j = r.json()

        return Action(j.get('function', {}), connector)



"""
Copyright 2019 Cognitive Scale, Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pickle
import json
from typing import Dict
from pathlib import Path
from datetime import datetime
from abc import ABC, abstractmethod
from .utils import get_logger
from .exceptions import ConfigurationException
from .camel import CamelResource
from .pipeline import Pipeline
from .pipeline_loader import PipelineLoader
from .properties import PropertyManager
from .serviceconnector import _Client
from .utils import raise_for_status_with_detail

log = get_logger(__name__)


class DatasetsClient(_Client):
    """
    A client used to manage datasets.
    """
    URIs = {'datasets': 'datasets',
            'content':  'content'}

    def list_datasets(self):
        """
        Get a list of all datasets.
        """
        # TODO: Use pagination?
        uri = self.URIs['datasets']
        r = self._serviceconnector.request('GET', uri)
        raise_for_status_with_detail(r)
        return r.json().get('datasets', [])

    def save_dataset(self, dataset: Dict[str, object]):
        """
        Saves a dataset.

        :param dataset: A Cortex dataset as dict.
        """
        uri = self.URIs['datasets']
        body_s = json.dumps(dataset)
        headers = {'Content-Type': 'application/json'}
        r = self._serviceconnector.request('POST', uri, body_s, headers)
        raise_for_status_with_detail(r)
        return r.json()

    def get_dataframe(self, dataset_name: str):
        """
        Gets data from a dataset as a dataframe.

        :param dataset_name: The name of the dataset to pull data.
        :return: A dataframe dictionary
        """
        uri = '/'.join([self.URIs['datasets'], dataset_name, 'dataframe'])
        r = self._serviceconnector.request('GET', uri)
        raise_for_status_with_detail(r)
        return r.json()

    def get_stream(self, stream_name: str):
        """
        Gets a dataset as a stream.
        """
        uri = '/'.join([self.URIs['datasets'], stream_name, 'stream'])
        r = self._serviceconnector.request('GET', uri, stream=True)
        raise_for_status_with_detail(r)
        return r.raw

    def post_stream(self, stream_name, data):
        uri = '/'.join([self.URIs['datasets'], stream_name, 'stream'])
        headers = {"Content-Type": "application/json-lines"}
        r = self._serviceconnector.request('POST', uri, data, headers)
        print(r.text)
        raise_for_status_with_detail(r)
        return r.json()

    def get_pipeline(self, dataset_name: str, pipeline_name: str):
        uri = '/'.join([self.URIs['datasets'], dataset_name, 'pipelines', pipeline_name])
        r = self._serviceconnector.request('GET', uri)
        raise_for_status_with_detail(r)
        return r.json()


class _DatasetPipelineLoader(PipelineLoader):

    def __init__(self, ds):
        super().__init__()
        self.ds = ds

    def add_pipeline(self, name, pipeline):
        super().add_pipeline(name, pipeline)
        self.ds._add_pipeline(name, pipeline)

    def get_pipeline(self, name):
        try:
            p = self.ds._client.get_pipeline(self.ds.name, name)
            pipeline = Pipeline.load(p, self)
            self.add_pipeline(name, pipeline)
            return pipeline
        except:
            return super().get_pipeline(name)

    def remove_pipeline(self, name):
        super().remove_pipeline(name)
        self.ds._remove_pipeline(name)


class AbstractDataset(ABC):
    """
    Abstract base class for datasets.
    """
    @abstractmethod
    def get_dataframe(self):
        raise NotImplementedError()

    @abstractmethod
    def get_stream(self):
        raise NotImplementedError()

    @abstractmethod
    def as_pandas(self):
        raise NotImplementedError()

    def visuals(self, df=None, figsize=(18, 9)):
        if df is None:
            df = self.as_pandas()
        try:
            from .viz import Viz
            return Viz(df, figsize)
        except ImportError:
            raise ConfigurationException('Please install the "viz" extras to use this method')

    @abstractmethod
    def save(self):
        raise NotImplementedError()

    @abstractmethod
    def to_camel(self):
        raise NotImplementedError()

    @abstractmethod
    def pipeline(self, name, clear_cache=False, depends=None):
        raise NotImplementedError()


class Dataset(AbstractDataset, CamelResource):

    """
    Defines the data and query parameters for accessing inputs to
    a managed content file or database connection.
    """

    def __init__(self, ds, client: DatasetsClient):
        super().__init__(ds, read_only=False)
        self._client = client
        self._work_dir = Path.cwd() / 'datasets' / self.name
        self._pipeline_loader = _DatasetPipelineLoader(self)

        # Cache all configured pipelines into our Pipeline loader
        if self.pipelines:
            for name in self.pipelines.keys():
                self._pipeline_loader.get_pipeline(name)
        else:
            self.pipelines = {}

    @staticmethod
    def get_dataset(name, client: DatasetsClient):
        """
        Gets a dataset.

        :param name: The name of the dataset to retrieve.
        :param client: The client instance to use.
        :return: A dataset object.
        """
        uri = '/'.join(['datasets', name])
        log.debug('Getting dataset using URI: %s' % uri)
        r = client._serviceconnector.request('GET', uri)
        raise_for_status_with_detail(r)

        return Dataset(r.json(), client)

    def get_dataframe(self):
        """
        Gets the dataframe for the dataset.
        """
        return self._client.get_dataframe('{}:{}'.format(self.name, self.version))

    def get_stream(self):
        """
        Streams the data coming from the dataset.
        """
        return self._client.get_stream('{}:{}'.format(self.name, self.version))

    def as_pandas(self):
        """
        Gets a pandas dataframe for the dataset.
        """
        df = self.get_dataframe()
        columns = df.get('columns')
        values = df.get('values')

        try:
            import pandas as pd
            return pd.DataFrame(values, columns=columns)
        except ImportError:
            log.warn('Pandas is not installed, please run `pip install pandas` or equivalent in your environment')
            return {'columns': columns, 'values': values}

    def save(self):
        try:
            from cortex_builders import DatasetBuilder
            b = DatasetBuilder(self.name, self._client, self.camel)
            return b.from_dataset(self).build()
        except ImportError:
            raise ConfigurationException('The cortex-python-builders library is required to save Datasets')

    def to_camel(self, camel='1.0.0'):
        ds = {
            'camel': camel,
            'name': self.name,
            'title': self.title,
            'description': self.description,
            'parameters': self.parameters,
            'pipelines': {key: pipeline.to_camel(camel) for key, pipeline in self.pipelines.items()}
        }

        if camel == '1.0.0':
            ds['connectionName'] = self.connectionName
            ds['connectionQuery'] = self.connectionQuery
        else:
            ds['connections'] = self.connections

        return ds

    def _add_pipeline(self, name, pipeline):
        self.pipelines[name] = pipeline

    def _remove_pipeline(self, name):
        if name in self.pipelines:
            del self.pipelines[name]

    def pipeline(self, name, clear_cache=False, depends=None):
        """
        Gets a pipeline for the dataset.

        :param name: name of the pipeline
        :param clear_cache: a flag to indicate whether previous results for the pipeline should be cleared
        :param depends: a list of pipeline names upon which this pipeline depends
        """
        p = self._pipeline_loader.get_pipeline(name)
        if clear_cache:
            # TODO clear any caches - e.g. for local mode
            pass

        if depends is not None:
            for d in depends:
                p.add_dependency(self._pipeline_loader.get_pipeline(d))

        return p


class LocalDataset(AbstractDataset):
    """
    References datasets that are external to Cortex.
    """

    config_file = 'config.yml'
    root_key = 'dataset'
    dir_cortex = '.cortex'
    dir_local = 'local'
    dir_data = 'data'
    dir_datasets = 'datasets'
    camel = '1.0.0'
    environment_id = 'cortex/local'

    def __init__(self, name, basedir=None):
        if basedir:
            self._basedir = Path(basedir)
        else:
            self._basedir = Path.home() / self.dir_cortex

        self._work_dir = self._basedir / self.dir_local / self.dir_datasets / name
        self._work_dir.mkdir(parents=True, exist_ok=True)
        self._data_dir = Path(self._work_dir / self.dir_data)
        self._data_dir.mkdir(parents=True, exist_ok=True)
        self._pipeline_loader = PipelineLoader()

        # Initialize config
        pm = PropertyManager()
        try:
            # Load local dataset configuration
            pm.load(str(self._work_dir / self.config_file))

            # Cache all configured pipelines into our Pipeline loader
            pipelines = pm.get('pipelines') or {}
            for p_name, p_doc in pipelines.items():
                Pipeline.load(p_doc, self._pipeline_loader)
        except FileNotFoundError:
            # Set initial configuration state
            pm.set('name', name)
            pm.set('camel', self.camel)
            pm.set('_environmentId', self.environment_id)
            pm.set('createdAt', str(datetime.now()))
            pm.set('_version', 0)

        self._config = pm

    @property
    def name(self):
        return self._config.get('name')

    @name.setter
    def name(self, title):
        self._config.set('name', title)

    @property
    def version(self):
        return self._config.get('_version')

    @property
    def title(self):
        return self._config.get('title')

    @title.setter
    def title(self, title):
        self._config.set('title', title)

    @property
    def description(self):
        return self._config.get('description')

    @description.setter
    def description(self, description):
        self._config.set('description', description)

    @property
    def parameters(self):
        return self._config.get('parameters') or []

    @parameters.setter
    def parameters(self, params):
        self._config.set('parameters', params)

    def save(self):
        # Save pipelines from the loader
        pipelines = self._pipeline_loader.dump(self.camel)
        for k, v in pipelines.items():
            self._config.set(self._config.join('pipelines', k), v)

        self._config.set('_version', self.version + 1)
        self._config.save(self._work_dir / self.config_file)

        return self

    def as_pandas(self):
        return self.get_dataframe()

    @property
    def data_dir(self):
        return self._data_dir

    @property
    def content_key(self):
        return self._config.get('content_key')

    @content_key.setter
    def content_key(self, content_key: str):
        if not content_key:
            return
        ext = content_key.split('.')[-1]
        self._config.set('content_key', content_key)
        self._config.set('content_type', ext)

    @property
    def content_type(self):
        return self._config.get('content_type')

    def get_dataframe(self):
        try:
            import pandas as pd

            data_file = str(self._data_dir / self.content_key)

            if self.content_type == 'csv':
                return pd.read_csv(data_file)
            elif self.content_type == 'json':
                return pd.read_json(data_file, orient='records', lines=True)
            else:
                with self.get_stream() as s:
                    return pickle.load(s)
        except ImportError:
            raise ImportError('Pandas is not installed, please run `pip install pandas` or equivalent in your environment')

    def get_stream(self):
        data_file = str(self._data_dir / self.content_key)
        return open(data_file, 'rb')

    def to_camel(self):
        pipelines = self._config.get('pipelines') or {}
        ds = {
            'camel': self.camel,
            'name': self.name,
            '_version': self.version,
            'title': self.title,
            'description': self.description,
            'parameters': self.parameters,
            'pipelines':  pipelines
        }

        return ds

    def pipeline(self, name, clear_cache=False, depends=None):
        pipelines = self._config.get('pipelines') or {}
        p_doc = pipelines.get(name)
        if p_doc:
            return Pipeline.load(p_doc, self._pipeline_loader)

        p = Pipeline(name, depends, self._pipeline_loader)
        # pipelines[name] = p.to_camel(self.camel)
        # self._config.set('pipelines', pipelines)
        self.save()

        return p
